from matplotlib.offsetbox import AnchoredOffsetbox, AuxTransformBox, VPacker,\
     TextArea, AnchoredText, DrawingArea, AnnotationBbox

from mpl_toolkits.axes_grid1.anchored_artists import \
         AnchoredDrawingArea, AnchoredAuxTransformBox, \
         AnchoredEllipse, AnchoredSizeBar
